/*
 * ts2vob - Convert MPEG Transport Stream to Program Stream for DVD-Video
 * Copyright (C) 2005 Martin Winken <martinwi@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 */

/* $Id: MPEGAudioStream.cc,v 1.2 2006/07/20 16:14:55 martinwi Exp $ */

#include <iostream>
#include <cmath>
#include <cstdlib>

#include "MPEGAudioStream.h"
#include "Cutter.h"
#include "TimeStamp.h"
#include "AudioAccessUnit.h"

const int MPEGAudioStream::SamplesPerFrame[3] = { 384, 1152, 1152 };

const int MPEGAudioStream::SlotsPerFrame[3] = { 12, 144, 144 };

const int MPEGAudioStream::BytesPerSlot[3] = { 4, 1, 1 };

const int MPEGAudioStream::BitRate[2][3][16] = {
    { // ISO/IEC 13818-3 lower sampling frequency extension
        { 0,32,48,56,64,80,96,112,128,144,160,176,192,224,256,0 },
        { 0,8,16,24,32,40,48,56,64,80,96,112,128,144,160,0 },
        { 0,8,16,24,32,40,48,56,64,80,96,112,128,144,160,0 }
    },
    { // ISO/IEC 11172-3 audio stream
        { 0,32,64,96,128,160,192,224,256,288,320,352,384,416,448,0 },
        { 0,32,48,56,64,80,96,112,128,160,192,224,256,320,384,0 },
        { 0,32,40,48,56,64,80,96,112,128,160,192,224,256,320,0 }
    }
};

const int MPEGAudioStream::SampleRate[2][4] = {
    { // ISO/IEC 13818-3 lower sampling frequency extension
        22050, 24000, 16000, 0
    },
    { // ISO/IEC 11172-3 audio stream
        44100, 48000, 32000, 0
    }
};

MPEGAudioStream::MPEGAudioStream(Cutter &out_) :
    out(out_), seekSynchronization(true), waitingForPTS(true),
    lastHeader(4, 0x00)
{
}

void MPEGAudioStream::newAAUcomplete()
{
    if (hasSavedTimeStamp()) {
        // AAU has PTS
        const TimeStamp savedPTS = getSavedPTS();
        nextSavedTimeStamp();

        if (waitingForPTS) {
            waitingForPTS = false;
        }
        else if (abs(savedPTS-PTS) > 5) {
            std::cerr << "\nMPEG-Audio-PTS-Error: expected " << PTS
                      << ", got " << savedPTS << std::endl;
        }
        PTS = savedPTS;
    }
    else if (waitingForPTS) {
        removeFromBuffer(length);
        return;
    }

    bytevec AAUdata;
    takeFromBuffer(length, AAUdata);
    AudioAccessUnit *aau = new AudioAccessUnit(AAUdata, PTS);
    out.takeAU(aau);
}

void MPEGAudioStream::discardAU()
{
    out.flush();
    waitingForPTS = true;
    seekSynchronization = true;
    clearBuffer();
}

inline bool MPEGAudioStream::isSyncWord(const uint8_t *ptr, bool lazy)
{
    return (ptr[0] == 0xff) && ((ptr[1] & Bitmask8(4,4)) == 0xf0) &&
        (lazy || ((getLayer(ptr) != 0x0) &&
                  (getSamplingFrequencyCode(ptr) != 0x3) &&
                  (getBitRateIndex(ptr) != 0x0)));
}

inline int MPEGAudioStream::getID(const uint8_t *ptr)
{
    return (ptr[1] >> 3) & Bitmask8(1);
}

inline int MPEGAudioStream::getLayer(const uint8_t *ptr)
{
    return (ptr[1] >> 1) & Bitmask8(2);
}

inline int MPEGAudioStream::getSamplingFrequencyCode(const uint8_t *ptr)
{
    return (ptr[2] >> 2) & Bitmask8(2);
}

inline int MPEGAudioStream::getBitRateIndex(const uint8_t *ptr)
{
    return ptr[2] >> 4;
}

inline int MPEGAudioStream::getPaddingBit(const uint8_t *ptr)
{
    return (ptr[2] >> 1) & Bitmask8(1);
}

unsigned int MPEGAudioStream::getFrameLength(const uint8_t *ptr)
{
    int layer = 3 - getLayer(ptr),
        bit_rate_index = getBitRateIndex(ptr),
        freq_code = getSamplingFrequencyCode(ptr),
        id = getID(ptr),
        padding = getPaddingBit(ptr);

    long int slots = SlotsPerFrame[layer]*
        BitRate[id][layer][bit_rate_index]*1000/SampleRate[id][freq_code];

    return (slots+padding)*BytesPerSlot[layer];
}

unsigned int MPEGAudioStream::getFrameDuration(const uint8_t *ptr)
{
    int layer = 3 - getLayer(ptr),
        freq_code = getSamplingFrequencyCode(ptr),
        id = getID(ptr);

    return unsigned(std::ceil(double(SamplesPerFrame[layer]*
                                     TicksPerSecond/
                                     SampleRate[id][freq_code])));
}

bool MPEGAudioStream::checkSynchronization(unsigned int offset, int count)
{
    if ((buffer.size() < (offset+4)) || !isSyncWord(&buffer[offset], false))
        return false;

    int length = getFrameLength(&buffer[offset]);
    if (length == 0)
        return false;
    if (count > 0) {
        return checkSynchronization(offset+length, count-1);
    }
    else {
        return true;
    }
}

bool MPEGAudioStream::trySynchronization()
{
    for (unsigned i = 0; (i+3) < buffer.size(); i++) {
        if (checkSynchronization(i, CheckSyncFrames)) {
            removeFromBuffer(i);
            return true;
        }
    }
    return false;
}

void MPEGAudioStream::buildAUs(bytevec::const_iterator insertPos)
{
    while (buffer.size() > 3) {
        if (!isSyncWord(&buffer.front(), true) && !seekSynchronization) {
            std::cerr << "MPEG-Audio: Sync lost" << std::endl;
            seekSynchronization = true;
        }

        if (seekSynchronization) {
            if (!trySynchronization())
                return;
            seekSynchronization = false;
        }

        bytevec header(buffer.begin(), buffer.begin()+4);
        header[2] &= ~Bitmask8(1); // clear private bit
        if (header != lastHeader) {
            lastHeader.swap(header);
            duration = getFrameDuration(&buffer.front());
            length = getFrameLength(&buffer.front());
        }

        if (length > buffer.size())
            break;

        newAAUcomplete();
        PTS += duration;
    }
}

void MPEGAudioStream::endOfStream()
{
    discardAU();
}
