/*
 * ts2vob - Convert MPEG Transport Stream to Program Stream for DVD-Video
 * Copyright (C) 2005 Martin Winken <martinwi@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 */

/* $Id: MPEGVideoStream.cc,v 1.2 2006/07/20 16:14:55 martinwi Exp $ */

#include <iostream>

#include "MPEGVideoStream.h"
#include "VideoAccessUnit.h"
#include "Cutter.h"

const int MPEGVideoStream::FrameDuration[] = {
    3753 * 4 + 3,   // 23,976 Hz
    3750 * 4,       // 24 Hz
    3600 * 4,       // 25 Hz
    3003 * 4,       // 29,97 Hz
    3000 * 4,       // 30 Hz
    1800 * 4,       // 50 Hz
    1501 * 4 + 2,   // 59,94 Hz
    1500 * 4        // 60 Hz
};

MPEGVideoStream::MPEGVideoStream(Cutter &out_, long video_rate) :
    out(out_),
    brokenLink(true), waitingForSeqHeader(true), PicFound(false),
    GOPStartIndex(-1), PICStartIndex(-1), SeqEndStartIndex(-1),
    state(Init),
    lastVAU(NULL), videoRate(video_rate),
    duration(-1), framesInGOP(0)
{
}

void MPEGVideoStream::newVAUcomplete(int size)
{
    if (size == 0)
        return;

    if (waitingForSeqHeader) {
        if (hasSavedTimeStamp())
            nextSavedTimeStamp();
        removeFromBuffer(size);
        GOPStartIndex = PICStartIndex = SeqEndStartIndex = -1;
        return;
    }

    if (isSequenceStart()) {
        int code = getFrameRateCode();
        if (code < 1 || code > 8) {
            std::cerr << "Invalid frame_rate_code: " << code << std::endl;
            exit(1);
        }
        duration = FrameDuration[code-1];
    }

    TimeStamp PTS, DTS;
    if (GOPStartIndex >= 0) {
        if (!hasSavedTimeStamp()) {
            std::cerr << "First VAU in GOP without PTS/DTS" << std::endl;
            exit(1);
        }
        PTS = getSavedPTS();
        DTS = getSavedDTS();
        nextSavedTimeStamp();

        framesInGOP = 0;
        firstPTSinGOP = PTS * 4 - getTemporalReference() * duration;
        firstDTSinGOP = DTS * 4;
    }
    else {
        if (duration <= 0) {
            std::cerr << "Frame rate not known" << std::endl;
            exit(1);
        }
        PTS = (firstPTSinGOP + getTemporalReference() * duration + 2) / 4;
        DTS = (firstDTSinGOP +            framesInGOP * duration + 2) / 4;
        if (hasSavedTimeStamp()) {
            //if (abs(PTS - getSavedPTS()) > 0)
            if (PTS != getSavedPTS())
                std::cerr << "Video-PTS-Error: computed=" << PTS
                          << ", coded=" << getSavedPTS()
                          << ", diff="  << PTS-getSavedPTS() << std::endl;
            //if (abs(DTS - getSavedDTS()) > 0)
            if (DTS != getSavedDTS())
                std::cerr << "Video-DTS-Error: computed=" << DTS
                          << ", coded=" << getSavedDTS()
                          << ", diff="  << DTS-getSavedDTS() << std::endl;
            PTS = getSavedPTS();
            DTS = getSavedDTS();
            nextSavedTimeStamp();           
        }
    }
    framesInGOP++;

    if (lastVAU != NULL) {
        out.takeAU(lastVAU);
    }

    bytevec VAUdata;
    takeFromBuffer(size, VAUdata);
    lastVAU = new VideoAccessUnit(VAUdata, PTS, DTS,
                                    GOPStartIndex, PICStartIndex,
                                    SeqEndStartIndex);
    GOPStartIndex = PICStartIndex = SeqEndStartIndex = -1;

    if (brokenLink) {
        lastVAU->setBrokenLink();
        brokenLink = false;
    }

    if (lastVAU->isSequenceStart() && (videoRate > 0))
        lastVAU->setBitRate(videoRate);
}

void MPEGVideoStream::discardAU()
{
    if (lastVAU != NULL) {
        lastVAU->writeSequenceEndCode();
        out.takeAU(lastVAU);
        out.flush();
        lastVAU = NULL;
    }
    waitingForSeqHeader = true;
    brokenLink = true;
}

void MPEGVideoStream::buildAUs(bytevec::const_iterator insertPos)
{
    bytevec::const_iterator iter = insertPos;
    while (iter != buffer.end()) {
        switch (state) {
            case Init:
                if (*iter == 0)
                    state = GotFirstZero;
                break;
            case GotFirstZero:
                if (*iter == 0)
                    state = GotSecondZero;
                else
                    state = Init;
                break;
            case GotSecondZero:
                if (*iter == 1)
                    state = GotOne;
                else if (*iter != 0)
                    state = Init;
                break;
            case GotOne:
                switch (*iter) {
                    case StartCode::SequenceHeaderCode:
                        newVAUcomplete(iter-buffer.begin()-3);
                        iter = buffer.begin()+3;
                        waitingForSeqHeader = false;
                        PicFound = false;
                        break;
                    case StartCode::GroupStartCode:
                        if (PicFound) {
                            newVAUcomplete(iter-buffer.begin()-3);
                            iter = buffer.begin()+3;
                            PicFound = false;
                        }
                        GOPStartIndex = iter-buffer.begin()-3;
                        break;
                    case StartCode::PictureStartCode:
                        if (PicFound) {
                            newVAUcomplete(iter-buffer.begin()-3);
                            iter = buffer.begin()+3;
                        }
                        PicFound = true;
                        PICStartIndex = iter-buffer.begin()-3;
                        break;
                    case StartCode::SequenceEndCode:
                        SeqEndStartIndex = iter-buffer.begin()-3;
                        newVAUcomplete(iter-buffer.begin()+1);
                        iter = buffer.begin();
                        waitingForSeqHeader = true;
                        state = Init;  // skip ++iter
                        continue;      // iter points already to next byte
                }
                state = Init;
                break;
        }
        ++iter;
    }
}

void MPEGVideoStream::endOfStream()
{
    discardAU();
}
