/*
 * ts2vob - Convert MPEG Transport Stream to Program Stream for DVD-Video
 * Copyright (C) 2005 Martin Winken <martinwi@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 */

/* $Id: PESPacket.h,v 1.2 2006/07/20 16:14:55 martinwi Exp $ */

#ifndef PESPACKET_H
#define PESPACKET_H

#include "ts2vob.h"
#include "TimeStamp.h"

class TimeStamp;

class PESPacket {
private:
    const uint8_t *packetData;
    int size;

    TimeStamp decodeTimeStamp(const uint8_t *ptr) const;

public:
    PESPacket() : packetData(NULL) {}
    void setData(const uint8_t *data, int datasize);

    enum StreamType {
        MPEGAudioStream,
        MPEGVideoStream,
        PrivateStream1,
        OtherStream
    };

    inline bool startCodeOk() const;
    inline uint8_t getStreamID() const;
    inline StreamType getStreamType() const;
    inline int getStreamNr() const;
    inline bool hasCRC() const;
    inline bool getCopyright() const;
    inline bool getOriginal() const;
    inline int getPTS_DTSFlags() const;
    inline bool hasPTS() const;
    inline bool hasDTS() const;
    inline TimeStamp getPTS() const;
    inline TimeStamp getDTS() const;
    inline int getHeaderDataLength() const;
    inline const uint8_t *getPayload() const;
    inline int getPayloadSize() const;
};

inline void PESPacket::setData(const uint8_t *data, int datasize)
{
    size = datasize;
    packetData = data;
}

inline bool PESPacket::startCodeOk() const
{
    return packetData[0] == 0 && packetData[1] == 0 &&
        packetData[2] == 1;
}

inline uint8_t PESPacket::getStreamID() const
{
    return packetData[3];
}

inline int PESPacket::getStreamNr() const
{
    uint8_t stream_id = getStreamID();

    if (IsVideoStreamID(stream_id))
        return stream_id & Bitmask8(4);
    else if (IsAudioStreamID(stream_id))
        return stream_id & Bitmask8(5);
    else
        return 0;
   
}

inline bool PESPacket::getCopyright() const
{
    return ((packetData[6] >> 1) & 1) == 1;
}

inline bool PESPacket::getOriginal() const
{
    return (packetData[6] & 1) == 1;
}

inline bool PESPacket::hasCRC() const
{
    return ((packetData[7] >> 1) & 1) == 1;
}

inline PESPacket::StreamType PESPacket::getStreamType() const
{
    int stream_id = getStreamID();

    if (IsVideoStreamID(stream_id))
        return MPEGVideoStream;
    else if (IsAudioStreamID(stream_id))
        return MPEGAudioStream;
    else if (stream_id == StreamID::PrivateStream1)
        return PrivateStream1;
    else
        return OtherStream;
}

inline int PESPacket::getPTS_DTSFlags() const
{
    return (packetData[7] >> 6) & Bitmask8(2);
}

inline bool PESPacket::hasPTS() const
{
    return (getPTS_DTSFlags() & 2) == 2;
}

inline bool PESPacket::hasDTS() const
{
    return (getPTS_DTSFlags() & 1) == 1;
}

inline TimeStamp PESPacket::decodeTimeStamp(const uint8_t *ptr) const
{
    return ((int64_t)(ptr[0] & Bitmask8(3, 1)) << 29) |
        ((int64_t)ptr[1] << 22) |
        ((int64_t)(ptr[2] & Bitmask8(7, 1)) << 14) |
        ((int64_t)ptr[3] << 7) |
        ((int64_t)ptr[4] >> 1);
}

inline TimeStamp PESPacket::getPTS() const
{
    return decodeTimeStamp(&packetData[9]);
}

inline TimeStamp PESPacket::getDTS() const
{
    return decodeTimeStamp(&packetData[14]);
}

inline int PESPacket::getHeaderDataLength() const
{
    return packetData[8];
}

inline const uint8_t *PESPacket::getPayload() const
{
    return &packetData[9+getHeaderDataLength()];
}

inline int PESPacket::getPayloadSize() const
{
    return size-9-getHeaderDataLength();
}

#endif
