/*
 * ts2vob - Convert MPEG Transport Stream to Program Stream for DVD-Video
 * Copyright (C) 2005 Martin Winken <martinwi@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 */

/* $Id: PIDHandler.cc,v 1.2 2006/07/20 16:14:55 martinwi Exp $ */

#include <iostream>

#include "PIDHandler.h"
#include "TSPacket.h"
#include "VOBStream.h"
#include "Cutter.h"
#include "Packetizer.h"
#include "MPEGVideoStream.h"
#include "MPEGAudioStream.h"
#include "AC3AudioStream.h"
#include "DummyStream.h"

void PIDHandler::takeTSPacket(const TSPacket &packet)
{
    if (!packet.hasPayload())
        return;

    int cc = packet.getCC();

#if 0
    if (cc == lastCC) {
        std::cerr << "PID " << packet.getPID() << ": Duplicate Packet"
                  << std::endl;
        return; // duplicate packet
    }
#endif

    if (packet.getTransportError()) {
        std::cerr << "PID " << packet.getPID() << ": Transport Error"
                  << std::endl;
        if (ES != NULL)
            ES->discardAU();
        lastCC = cc;
        return;
    }
    else if (!ccCorrect(cc)) {
        std::cerr << "PID " << packet.getPID() << ": previous CC="
                  << lastCC << ", current CC=" << cc << std::endl;
        if (ES != NULL)
            ES->discardAU();
    }
    lastCC = cc; 

    const uint8_t *payload = packet.getPayload();
    int payloadSize = packet.getPayloadSize();
    bool hasTimeStamp = false;
    TimeStamp PTS, DTS;

    if (packet.getUnitStart()) {
        pesp.setData(payload, payloadSize);
        payload = pesp.getPayload();
        payloadSize = pesp.getPayloadSize();

        int PTS_DTS_flags = pesp.getPTS_DTSFlags();
        if (PTS_DTS_flags & 2) {
            hasTimeStamp = true;
            PTS = pesp.getPTS();
            if (PTS_DTS_flags & 1)
                DTS = pesp.getDTS();
            else
                DTS = PTS;
        }

        if (ES == NULL) {
            switch (pesp.getStreamType()) {
                case PESPacket::MPEGVideoStream:
                    std::cerr << "PID " << packet.getPID()
                              << " is MPEG-Videostream " << pesp.getStreamNr()
                              << std::endl;
                    packetizer = new Packetizer(pesp.getStreamID(),
                                                vob, 2034, 232*1024);
                    cutter = new Cutter(*packetizer, startIter, endIter);
                    ES = new MPEGVideoStream(*cutter, videoRate);
                    break;
                case PESPacket::MPEGAudioStream:
                    std::cerr << "PID " << packet.getPID()
                              << " is MPEG-Audiostream " << pesp.getStreamNr()
                              << std::endl;
                    packetizer = new Packetizer(pesp.getStreamID(),
                                                vob, 2034, 4096);
                    cutter = new Cutter(*packetizer, startIter, endIter);
                    ES = new MPEGAudioStream(*cutter);
                    break;
                case PESPacket::PrivateStream1:
                    std::cerr << "PID " << packet.getPID()
                              << " is AC3-Audiostream " << packet.getPID() % 8
                              << std::endl;
                    // ETSI TS 101 154 recommends a main audio buffer size of
                    // 5696 bytes. The value 59392 is the DVD buffer size for
                    // private_stream_1 data.
                    packetizer = new Packetizer(pesp.getStreamID(),
//                                              vob, 2034, 59392,
                                                vob, 2034, 5696,
                                                true, packet.getPID() % 8);
                    cutter = new Cutter(*packetizer, startIter, endIter);
                    ES = new AC3AudioStream(*cutter);
                    break;
                default:
                    std::cerr << "PID " << packet.getPID()
                              << " gets ignored (Stream-ID 0x"
                              << std::hex << int(pesp.getStreamID())
                              << std::dec << ")" << std::endl;
                    ES = new DummyStream();
                    break;
            }
        }
    }

    if (ES != NULL && !ignore)
        ES->addData(payload, payloadSize, hasTimeStamp, PTS, DTS);
}

inline bool PIDHandler::ccCorrect(int cc)
{
    return (lastCC == -1) || ((lastCC+1)%16 == cc);
}

PIDHandler::~PIDHandler()
{
    if (ES != NULL)
        delete ES;
    if (cutter != NULL)
        delete cutter;
    if (packetizer != NULL)
        delete packetizer;
}

void PIDHandler::endOfStream()
{
    if (ES != NULL)
        ES->endOfStream();
}
