/*
 * ts2vob - Convert MPEG Transport Stream to Program Stream for DVD-Video
 * Copyright (C) 2005 Martin Winken <martinwi@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 */

/* $Id: Packetizer.cc,v 1.2 2006/07/20 16:14:55 martinwi Exp $ */

#include <iostream>
#include <algorithm>

#include "Packetizer.h"
#include "OutputPacket.h"
#include "AccessUnit.h"


Packetizer::Packetizer(uint8_t id, VOBStream &out, int packSiz,
                       int stdBufferSize,
                       bool ac3, uint8_t subid) :
    buffer(stdBufferSize), bufferSize(stdBufferSize),
    streamID(id), packetSize(packSiz),
    vob(out), totalAUBytes(0), firstPacket(true),
    writeAC3Header(ac3), AC3Id(subid)
{
}

void Packetizer::takeAU(AccessUnit *au)
{
    if (au->startNewPacketHere())
        flush();
    aus.push_back(au);
    totalAUBytes += au->getSize();
    writePackets(false);
}

void Packetizer::flush()
{
    writePackets(true);
}

void Packetizer::writePackets(bool paddingIfNeeded)
{
    OutputPacket *packet;
    while ((packet = getPacket(paddingIfNeeded)) != NULL) {
        vob.takePacket(packet);
    }
}

OutputPacket *Packetizer::getPacket(bool paddingIfNeeded)
{
    if (aus.empty())
        return NULL;
    bool writeTimeStamp = false, writeBufferSize = false;
    int bytesToFill = packetSize - 9;

    if (writeAC3Header) {
        bytesToFill -= 4;
    }

    if (firstPacket) {
        writeBufferSize = true;
        bytesToFill -= 3;
    }

    TimeStamp PTS, DTS;
    handleTimeStamps(bytesToFill, PTS, DTS, writeTimeStamp);

    int payloadSize = bytesToFill;
    if (totalAUBytes < bytesToFill) {
        if (!paddingIfNeeded)
            return NULL;
        payloadSize = totalAUBytes;
    }

    bytevec payload;
    if (writeAC3Header) {
        payload.resize(4);
        payloadSize += 4;
    }
    payload.reserve(payloadSize);

    TimeStamp maxSCR = aus.front()->getDTS();
    TimeStamp minSCR = firstPacket ?
        TimeStamp(maxSCR-TicksPerSecond) :
        buffer.getEarliestTime(payloadSize);
    bool hasAUEnd = false, wantsNV_PCK = false;
    int AUStart = 0, AUEnd = 0, AUStartCounter = 0;

    while ((bytesToFill > 0) && !aus.empty()) {
        AccessUnit *au = aus.front();
        const uint8_t *dataPtr = au->getRestData();
        int restSize = au->getRestSize();

        if (au->getBytesRead() == 0) {
            wantsNV_PCK = wantsNV_PCK || au->wantsNV_PCK();
            if (AUStartCounter == 0)
                AUStart = payload.size();
            AUStartCounter++;
        }

        minSCR = std::max(minSCR, TimeStamp(au->getDTS()-TicksPerSecond));
        if (restSize > bytesToFill) {
            /* AU doesn't fit into packet */
            payload.insert(payload.end(), dataPtr, &dataPtr[bytesToFill]);
            au->takeBytes(bytesToFill);
            totalAUBytes -= bytesToFill;
            buffer.store(bytesToFill, au->getDTS());
            bytesToFill = 0;
        }
        else {
            payload.insert(payload.end(), dataPtr, &dataPtr[restSize]);
            totalAUBytes -= restSize;
            buffer.store(restSize, au->getDTS());
            delete au;
            aus.pop_front();
            bytesToFill -= restSize;
            if (!hasAUEnd) {
                hasAUEnd = true;
                AUEnd = payload.size();
            }
        }
    }
    buffer.releaseUntil(minSCR);
    firstPacket = false;
    if (writeAC3Header) {
        payload[0] = AC3Id | 0x80;
        payload[1] = AUStartCounter;
        if (AUStartCounter) {
            AUStart -= 3;
            payload[2] = (AUStart >> 8) & 0xff;
            payload[3] = AUStart & 0xff;
        }
    }
    return new OutputPacket(streamID,
                            payload, bytesToFill,
                            minSCR, maxSCR,
                            PTS, DTS,
                            writeTimeStamp,
                            wantsNV_PCK,
                            writeBufferSize,
                            bufferSize,
                            hasAUEnd, AUEnd);
}

void Packetizer::handleTimeStamps(int &bytesToFill, TimeStamp &PTS, TimeStamp &DTS, bool &writeTimeStamp)
{
    AccessUnit &firstAUinPacket = *aus.front();

    PTS = firstAUinPacket.getPTS();
    DTS = firstAUinPacket.getDTS();
    if (firstAUinPacket.getBytesRead() == 0) {
        if (firstAUinPacket.writeTimeStamp()) {
            writeTimeStamp = true;
            if (PTS == DTS)
                bytesToFill -= 5;       // only PTS
            else
                bytesToFill -= 10;      // PTS and DTS
        }
    }
    else if ((aus.size() > 1) &&
             (bytesToFill > firstAUinPacket.getRestSize())) {
        AccessUnit &secondAUinPacket = *aus[1];
        PTS = secondAUinPacket.getPTS();
        DTS = secondAUinPacket.getDTS();
        if (secondAUinPacket.writeTimeStamp()) {
            int rest = bytesToFill;
            if (PTS == DTS)
                rest -= 5;
            else
                rest -= 10;
            if (rest > firstAUinPacket.getRestSize()) {
                writeTimeStamp = true;
                bytesToFill = rest;
            }
        }
    }
}
