/*
 * ts2vob - Convert MPEG Transport Stream to Program Stream for DVD-Video
 * Copyright (C) 2005 Martin Winken <martinwi@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 */

/* $Id: VOBStream.cc,v 1.2 2006/07/20 16:14:55 martinwi Exp $ */

#include <iostream>
#include <algorithm>
#include <cmath>

#include "VOBStream.h"
#include "OutputPacket.h"

VOBStream::VOBStream(std::ostream &out_, unsigned long mux_rate) :
    initializing(true),
    out(out_), SCR(0), muxRate(mux_rate/50),
    packSCRInkrement(int64_t(ceil((double)
                                  PackLength/mux_rate*SCRTicksPerSecond))),
    maxAudioBufferSize(0), maxVideoBufferSize(0)
{
}

void VOBStream::takePacket(const OutputPacket *packet)
{
    uint8_t id = packet->getStreamID();

    if (IsVideoStreamID(id)) {
        maxVideoBufferSize = std::max(maxVideoBufferSize,
                                         packet->getBufferSize());
        videoStreams.insert(id);
    }
    else if (IsAudioStreamID(id)) {
        maxAudioBufferSize = std::max(maxAudioBufferSize,
                                         packet->getBufferSize());
        audioStreams.insert(id);
    }

    if (packet->isAUEnd()) {
        writePackets(false, packet->minSCRBase()-2*TicksPerSecond);
    }
    packetQueues[id].push(packet);
}

void VOBStream::endOfStream()
{
    writePackets(true, 0);
}

// The following method schedules the packets using a "non-idling
// non-preemptive earliest deadline first" (NINP-EDF) policy.
void VOBStream::writePackets(bool flush, const TimeStamp &limit)
{
    const OutputPacket *packet;
    if (initializing) {
        if ((packet = firstPacket(packetQueues, flush, limit)) == NULL)
            return;
        zeroTime = packet->minSCRBase();
    }
    else {
        if ((packet = nextPacket(packetQueues, flush, limit)) == NULL)
            return;
    }

    do {
        SCR = computeNextSCR(*packet);
        if (initializing || isNV_PCKDue(*packet)) {
            std::cerr << '.';
            writeNV_PCK();
            DTSlastNV_PCK = packet->getDTS();
            initializing = false;
        }
        writePack(*packet);
        delete packet;
    }
    while ((packet = nextPacket(packetQueues, flush, limit)) != NULL);
}

const OutputPacket *VOBStream::firstPacket(QueueMap &queues,
                                            bool flush, const TimeStamp limit)
{
    return getPacket(queues, flush, limit, QueueEarlierReady());
}

const OutputPacket *VOBStream::nextPacket(QueueMap &queues,
                                     bool flush, const TimeStamp limit)
{
    TimeStamp now = SCR/300 + zeroTime;
    return getPacket(queues, flush, limit, QueueMoreUrgent(now));
}

int64_t VOBStream::computeNextSCR(const OutputPacket &packet)
{
    TimeStamp now = SCR/300 + zeroTime;
    int64_t newSCR = SCR;
    if (packet.minSCRBase() > now) {
        TimeStamp newSCRBase = packet.minSCRBase() - zeroTime;
        newSCR = newSCRBase * 300;
    }
    return newSCR;
}

bool VOBStream::isNV_PCKDue(const OutputPacket &packet) const
{
    const TimeStamp & DTS = packet.getDTS();

    return (videoStreams.empty() &&
            (DTS > TimeStamp(DTSlastNV_PCK+TicksPerSecond))) ||
        packet.wantsNV_PCK();

#if 0
    return (videoStreams.empty() || IsVideoStreamID(packet.getStreamID()))
            && (packet.wantsNV_PCK() ||
                (DTS > TimeStamp(DTSlastNV_PCK+TicksPerSecond)));
    return (videoStreams.empty() || IsVideoStreamID(packet.getStreamID()))
            && ((packet.wantsNV_PCK() &&
                 (DTS > TimeStamp(DTSlastNV_PCK+minTicksNV_PCK))) ||
                (DTS > TimeStamp(DTSlastNV_PCK+TicksPerSecond)));
    return (packet.wantsNV_PCK() &&
            (DTS > TimeStamp(DTSlastNV_PCK+minTicksNV_PCK))) ||
        (DTS > TimeStamp(DTSlastNV_PCK+TicksPerSecond));
#endif
}

template<class Compare> const OutputPacket *VOBStream::getPacket(
    QueueMap &queues, bool flush, const TimeStamp limit,
    Compare comp)
{
    if (queues.empty())
        return NULL;

    QueueMap::iterator i = min_element(queues.begin(), queues.end(), comp);
    PacketQueue &pQueue = i->second;
    const OutputPacket *packet = pQueue.front();

    if (!flush && (packet->minSCRBase() > limit))
        return NULL;

    pQueue.pop();
    if (pQueue.empty()) {
        if (!flush)
            std::cerr << "WARNING: Buffer underrun on stream "
                      << std::hex << int(i->first)
                      << std::dec << std::endl;
        queues.erase(i);
    }
    return packet;
}

// ----------------------------------------------------------------------
// The following methods write the actual output bytes to the VOB file.

void VOBStream::writeNV_PCK()
{
    writePackHeader();
    writeSystemHeader();
    writePCIPacket();
    writeDSIPacket();
    SCR += packSCRInkrement;
}

void VOBStream::writeSystemHeader()
{
    bytevec header(12, 0x00);
    
    header.reserve(24);
    header[2] = 0x01;
    header[3] = 0xBB;

    header[5] = 18; // 18 bytes follow
    
    header[6] = 0x80 | ((muxRate >> 15) & Bitmask8(7, 1));
    header[7] = (muxRate >> 7) & Bitmask8(8);
    header[8] = ((muxRate & Bitmask8(7)) << 1) | 0x01;

    header[9] = (audioStreams.size() & Bitmask8(6)) << 2; // audio_bound
    header[10] = 0xE0 | (videoStreams.size() & Bitmask8(5)); // video_bound
    header[11] = 0x7F;

    // buffer sizes follow
    header.push_back(0xB9); // all video streams
    writeBufferInfo(header, true, maxVideoBufferSize > 0 ?
                       maxVideoBufferSize : 237568, 0x3);
    
    header.push_back(0xB8); // all audio streams
    writeBufferInfo(header, false, maxAudioBufferSize > 0 ?
                       maxAudioBufferSize : 4096, 0x3);

    header.push_back(StreamID::PrivateStream1);
    writeBufferInfo(header, true, 59392, 0x3);

    header.push_back(StreamID::PrivateStream2);
    writeBufferInfo(header, true, 2048, 0x3);

    writeBytes(&header.front(), header.size());
}

void VOBStream::writePCIPacket()
{
    bytevec packet(986, 0x00);

    packet[2] = 0x01;
    packet[3] = StreamID::PrivateStream2;
    packet[4] = 0x03; // PES_packet_length = 0x03DE (980)
    packet[5] = 0xD4;

    writeBytes(&packet.front(), packet.size());
}

void VOBStream::writeDSIPacket()
{
    bytevec packet(1024, 0x00);

    packet[2] = 0x01;
    packet[3] = StreamID::PrivateStream2;
    packet[4] = 0x03; // PES_packet_length = 0x03FA (1018)
    packet[5] = 0xFA;

    packet[6] = 0x01; // Sub-Stream ID = 0x01
    writeBytes(&packet.front(), packet.size());
}

void VOBStream::writePack(const OutputPacket &packet)
{
    writePackHeader();
    writePacket(packet);
    SCR += packSCRInkrement;
}

void VOBStream::writePackHeader()
{
    bytevec data(14, 0x00);
    TimeStamp SCRBase = SCR/300;
    int SCRExtension = SCR%300;

    data[2] = 0x01;
    data[3] = 0xBA;

    data[4] = 0x44 |
        ((SCRBase >> 27) & Bitmask8(3,3)) |
        ((SCRBase >> 28) & Bitmask8(2));
    data[5] = (SCRBase >> 20) & Bitmask8(8);
    data[6] = ((SCRBase >> 12) & Bitmask8(5, 3)) |
        0x04 |
        (SCRBase >> 13) & Bitmask8(2);
    data[7] = (SCRBase >> 5) & Bitmask8(8);
    data[8] = ((SCRBase << 3) & Bitmask8(5, 3)) |
        0x04 |
        ((SCRExtension >> 7) & Bitmask8(8));
    data[9] = ((SCRExtension << 1) & Bitmask8(7, 1)) | 0x01;

    data[10] = (muxRate >> 14) & Bitmask8(8);
    data[11] = (muxRate >> 6) & Bitmask8(8);
    data[12] = ((muxRate & Bitmask8(6)) << 2) | 0x03 ;
    data[13] = 0xF8; // no stuffing_bytes

    writeBytes(&data.front(), data.size());
}

void VOBStream::writePacket(const OutputPacket &packet)
{
    writePacketHeader(packet);
    writeBytes(packet.getData(), packet.getSize());
    
    if (packet.getPaddingSize() > MaxStuffingbytes)
        writePaddingPacket(packet.getPaddingSize());
}

void VOBStream::writePaddingPacket(int size)
{
    bytevec data(size, 0xff);
    int packet_length = size - 6;

    data[0] = data[1] = 0x00;
    data[2] = 0x01;
    data[3] = StreamID::PaddingStream;
    data[4] = (packet_length >> 8) & Bitmask8(8);
    data[5] = packet_length & Bitmask8(8);
    writeBytes(&data.front(), size);
}

void VOBStream::writePacketHeader(const OutputPacket &packet)
{
    bytevec header(9, 0x00);

    header[2] = 0x01;
    header[3] = packet.getStreamID();
    
    header[6] = 0x81; // set original bit
    
    if (packet.writeTimeStamp()) {
        const TimeStamp & PTS = packet.getPTS();
        const TimeStamp & DTS = packet.getDTS();
        if (PTS == DTS) { // only PTS
            writeTimeStamp(header, PTS, 0x2);
            header[7] |= 0x80; // set PTS_DTS_flags
        }
        else { // PTS and DTS
            writeTimeStamp(header, PTS, 0x3);
            writeTimeStamp(header, DTS, 0x1);
            header[7] |= 0xC0; // set PTS_DTS_flags
        }
    }

    if (packet.writeBufferSize()) {
        header[7] |= 0x01; // set PES_extension_flag
        header.push_back(0x1E); // set P-STD_buffer_flag
        writeBufferInfo(header, !IsAudioStreamID(packet.getStreamID()),
                           packet.getBufferSize(), 0x1);
    }

    if (packet.getPaddingSize() <= MaxStuffingbytes) {
        header.insert(header.end(), packet.getPaddingSize(), 0xFF);
    }

    int packet_length = header.size() + packet.getSize() - 6;

    // set PES_packet_length
    header[4] = (packet_length >> 8) & Bitmask8(8);
    header[5] = packet_length & Bitmask8(8);

    header[8] = header.size() - 9; // set PES_header_data_length
    writeBytes(&header.front(), header.size());
}

void VOBStream::writeTimeStamp(bytevec &data, const TimeStamp &ts,
                                  int prefix)
{
    TimeStamp t = ts - zeroTime;
    uint8_t bytes[5];

    bytes[0] = ((prefix & Bitmask8(4)) << 4) |
        ((t >> 29) & Bitmask8(3, 1)) | 1;
    bytes[1] = (t >> 22) & Bitmask8(8);
    bytes[2] = ((t >> 14) & Bitmask8(7,1)) | 1;
    bytes[3] = (t >> 7) & Bitmask8(8);
    bytes[4] = ((t << 1) & Bitmask8(7,1)) | 1;
    data.insert(data.end(), bytes, &bytes[5]);
}

void VOBStream::writeBufferInfo(bytevec &data, bool scale, int bytes,
                                   int prefix)
{
    int size = bytes / (scale ? 1024 : 128);

    data.push_back((prefix & Bitmask8(2) << 6) |
                   scale ? 0x20 : 0x00 |
                   ((size >> 8) & Bitmask8(5)));
    data.push_back(size & Bitmask8(8));
}

inline void VOBStream::writeBytes(const uint8_t *data, int size)
{
    out.write((const char *)data, size);
}
