/*
 * ts2vob - Convert MPEG Transport Stream to Program Stream for DVD-Video
 * Copyright (C) 2005 Martin Winken <martinwi@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 */

/* $Id: VOBStream.h,v 1.3 2006/07/20 16:14:55 martinwi Exp $ */

#ifndef VOBSTREAM_H
#define VOBSTREAM_H

#include <list>
#include <queue>
#include <set>
#include <map>

#include "ts2vob.h"
#include "TimeStamp.h"
#include "OutputPacket.h"

class VOBStream {
private:
    enum {
        minTicksNV_PCK = 36000, // 0.4 sec
        MaxStuffingbytes = 32,
        PackLength = 2048,
    };

    class QueueMoreUrgent;
    class QueueEarlierReady;

    typedef std::queue<const OutputPacket*> PacketQueue;
    typedef std::map<uint8_t, PacketQueue> QueueMap;
    QueueMap packetQueues;

    bool initializing;
    std::ostream &out;
    TimeStamp zeroTime;
    int64_t SCR;
    TimeStamp DTSlastNV_PCK;
    unsigned long muxRate;
    int64_t packSCRInkrement;

    int maxAudioBufferSize, maxVideoBufferSize;
    std::set<uint8_t> audioStreams, videoStreams;

    // prevent copy and assignment of objects
    VOBStream(const VOBStream &);
    VOBStream & operator= (const VOBStream &);

    // methods used for scheduling
    void writePackets(bool flush, const TimeStamp &limit);
    const OutputPacket *firstPacket(QueueMap &queues,
                                     bool flush, const TimeStamp limit);
    const OutputPacket *nextPacket(QueueMap &queues,
                                    bool flush, const TimeStamp limit);
    int64_t computeNextSCR(const OutputPacket &packet);
    bool isNV_PCKDue(const OutputPacket &packet) const;
    template<class Compare> const OutputPacket *getPacket(
        QueueMap &queues, bool flush, const TimeStamp limit, Compare Comp);

    // methods used for writing the actual output bytes
    void writeNV_PCK();
    void writePCIPacket();
    void writeDSIPacket();
    void writePack(const OutputPacket &packet);
    void writePackHeader();
    void writeSystemHeader();
    void writePacket(const OutputPacket &packet);
    void writePacketHeader(const OutputPacket &packet);
    void writePaddingPacket(int size);
    void writeTimeStamp(bytevec &data, const TimeStamp &t,
                           int prefix);
    void writeBufferInfo(bytevec &data, bool scale, int size,
                            int prefix);
    void writeBytes(const uint8_t *data, int size);

public:
    VOBStream(std::ostream &out_, unsigned long mux_rate);
    void takePacket(const OutputPacket *packet);
    void endOfStream();
};

class VOBStream::QueueMoreUrgent {
private:
    TimeStamp now;

public:
    QueueMoreUrgent (const TimeStamp &l) : now(l) {}

    // returns true if queue a is more due than queue b, false otherwise
    // note: a and b must not be empty
    bool operator() (const QueueMap::value_type &a,
                     const QueueMap::value_type &b) const
        {
            const OutputPacket &pa = *a.second.front(),
                &pb = *b.second.front();
            if (pa.minSCRBase() <= now)
                // pa allowed to be scheduled
                return ((pb.minSCRBase() > now) || // pb not yet ready
                        (pa.maxSCRBase() <= pb.maxSCRBase())); // pa more due
            else if (pb.minSCRBase() <= now)
                // schedule pb
                return false;
            else
                // idle at the moment, check who is next
                return (pa.minSCRBase() <= pb.minSCRBase());
        }
};

class VOBStream::QueueEarlierReady {
public:
    // returns true if queue a is more due than queue b, false otherwise
    // note: a and b must not be empty
    bool operator() (const QueueMap::value_type &a,
                     const QueueMap::value_type &b) const
        {
            const OutputPacket &pa = *a.second.front(),
                &pb = *b.second.front();
            return (pa.minSCRBase() <= pb.minSCRBase());
        }
};


#endif
