/*
 * ts2vob - Convert MPEG Transport Stream to Program Stream for DVD-Video
 * Copyright (C) 2005 Martin Winken <martinwi@users.sourceforge.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 */

/* $Id: ts2vob.cc,v 1.3 2006/07/20 16:14:55 martinwi Exp $ */

#include <iostream>
#include <map>
#include <csignal>
#include <cstring>
#include <cstdlib>

#include "ts2vob.h"
#include "TransportStream.h"
#include "TSPacket.h"
#include "VOBStream.h"
#include "PIDHandler.h"
#include "Cut.h"

namespace {
    sig_atomic_t NoInterruptReceived = 1;
}

void SignalHandler(int sig)
{
    NoInterruptReceived = 0;
}

void InstallSignalHandler()
{
    struct sigaction sa;

    sa.sa_handler = SignalHandler;
    sigemptyset(&sa.sa_mask);
    sa.sa_flags = 0;
    sigaction(SIGINT, &sa, NULL);
    sigaction(SIGTERM, &sa, NULL);
}

void Usage(const char *progname)
{
    std::cerr << "Usage: " << progname
              << " [-i PID] [-n] [-t StartDTS-EndDTS] [-v Videobitrate]"
              << std::endl;
    exit(1);
}

int main(int argc, char *argv[])
{
    int opt;
    long video_bit_rate = 0;
    bool ignore = false;
    typedef std::map<int, PIDHandler*> pidmap;
    std::set<int> scrambled_pids;
    pidmap pids;
    CutDeque cuts;

    VOBStream vob(std::cout, 1260000);

    while ((opt = getopt(argc, argv, "i:nt:v:")) != -1) {
        switch (opt) {
            case 'i':  {
                char *token = std::strtok(optarg, ",");
                do {
                    int pid = std::strtol(token, NULL, 0);
                    if (pids.find(pid) == pids.end()) {
                        PIDHandler *pidhandler = new PIDHandler(vob, true,
                                                            video_bit_rate/400,
                                                            cuts.begin(),
                                                            cuts.end());
                        pids.insert(std::make_pair(pid, pidhandler));
                        std::cerr << "PID " << pid << " ("
                                  << std::hex << std::showbase << pid
                                  << std::dec << ") gets ignored" << std::endl;
                    }
                }
                while ((token = std::strtok(NULL, ",")) != NULL);
                }
                break;
            case 'n':
                ignore = true;
                break;
            case 't': {
                char *token = std::strtok(optarg, ",");
                do {
                    char *ptr;
                    if ((ptr = std::strchr(token, '-')) == NULL)
                        Usage(argv[0]);
                    *ptr = '\0'; ptr++;

                    TimeStamp start = std::strtoll(token, NULL, 0);
                    TimeStamp end = std::strtoll(ptr, NULL, 0);
                    cuts.push_back(Cut(start, end));
                }
                while ((token = std::strtok(NULL, ",")) != NULL);
                }
                break;
            case 'v':
                video_bit_rate = atol(optarg);
                break;
            case '?':
                Usage(argv[0]);
        }
    }

    InstallSignalHandler();

    TransportStream ts(std::cin);
    if (!ts.seekSync()) {
        std::cerr << "No TS." << std::endl;
        return 1;
    }

    TSPacket p;
    unsigned packnr = 0;

    while (ts.getPacket(p) && NoInterruptReceived) {
        packnr++;
        if (!p.syncOk()) {
            std::cerr << "SYNC-ERR in TS-Packet " << packnr << std::endl;
            continue;
        }
        int pid = p.getPID();
        if (p.isScrambled()) {
            if (scrambled_pids.find(pid) == scrambled_pids.end()) {
                std::cerr << "Ignoring scrambled TS-Packets (PID "
                          << pid << ")" << std::endl;
                scrambled_pids.insert(pid);
            }
            continue;
        }

        PIDHandler *pidhandler;
        pidmap::iterator i;
        if ((i = pids.find(pid)) == pids.end()) {
            pidhandler = new PIDHandler(vob, ignore,
                                        video_bit_rate/400,
                                        cuts.begin(),
                                        cuts.end());
            pids.insert(std::make_pair(pid, pidhandler));
        }
        else {
            pidhandler = i->second;
        }
        pidhandler->takeTSPacket(p);
    }

    for (pidmap::iterator i = pids.begin(); i != pids.end(); ++i) {
        PIDHandler *pidhandler = i->second;
        pidhandler->endOfStream();
        delete pidhandler;
    }
    vob.endOfStream();

    std::cerr << "DONE\n";
    return 0;
}
